<!doctype html><html lang="en"><head><meta charset="utf-8"/><meta name="viewport" content="width=device-width,initial-scale=1"/><link rel="manifest" href="/manifest.json"/><link rel="icon" type="image/png" href="{{favicon}}"/><link rel="apple-touch-icon" href="/favicon.png"/><title>{{title}}</title><script>{%- if config -%}
    window.taipyConfig={{config|tojson|safe}};{% endif %}</script><script defer="defer" src="/taipy.34f9e1d43662ea114715.js"></script></head><body><link rel="stylesheet" href="https://fonts.googleapis.com/css?family=Roboto:300,400,500,700&display=swap"/><link rel="stylesheet" href="{{app_css}}"/><noscript>You need to enable JavaScript to run this app.</noscript><div id="root" {%- if root_margin -%} style="margin:{{root_margin}};" {% endif %}></div>{%- if watermark -%}<span class="taipy-watermark" style="position:fixed;bottom:1em;opacity:0.4;color:gray;left:50%;transform:translateX(-50%);font-size:0.7em;">{{watermark}}</span>{% endif %}</body></html>